from .utils import Dataset
from .utils import Selection
from .utils import Ntuple
from .utils import Cut
from .utils import Weight
from .utils import Action
from .utils import Count
from .utils import Histogram
from .utils import Variation

from ROOT import gROOT
gROOT.SetBatch(True)
from ROOT import TFile

import os
import re
import json
import itertools

import logging
logger = logging.getLogger(__name__)



def dataset_from_artusoutput(
        dataset_name,
        file_names,
        folder,
        files_base_directory,
        friends_base_directories):
    """Create a Dataset object from a list containing the names
    of the ROOT files (e.g. [root_file1, root_file2, (...)]):
        ntuple1: /file_base_dir/root_file1/folder/ntuple
            friend1: /friend1_base_dir/root_file1/folder/ntuple
            friend2: /friend2_base_dir/root_file1/folder/ntuple
        ntuple2: /file_base_dir/root_file2/folder/ntuple
            friend1: /friend1_base_dir/root_file2/folder/ntuple
            friend2: /friend2_base_dir/root_file2/folder/ntuple
        ntuple3: /file_base_dir/root_file3/folder/ntuple
            friend1: /friend1_base_dir/root_file3/folder/ntuple
            friend2: /friend2_base_dir/root_file3/folder/ntuple
        (...)

    Args:
        dataset_name (str): Name of the dataset
        file_names (list): List containing the names of the .root
            files
        folder (str): Name of the TDirectoryFile in each .root file
        files_base_directory (str): Path to the files base directory (directories)
        friends_base_directories (str, list): List of paths to
            the friends base directory (directories)

    Returns:
        dataset (Dataset): Dataset object containing TTrees
    """
    def get_full_tree_name(folder, path_to_root_file, tree_name):
        root_file = TFile(path_to_root_file)
        if root_file.IsZombie():
            raise FileNotFoundError('File {} does not exist, abort'.format(path_to_root_file))
        if folder not in root_file.GetListOfKeys():
            raise NameError('Folder {} does not exist in {}\n'.format(folder, path_to_root_file))
        root_file.Close()
        full_tree_name = '/'.join([folder, tree_name])
        return full_tree_name

    def add_tagged_friends(friends):
        """ Tag friends with the name of the different directories
        in the artus name scheme, e.g.:
        /common_path/MELA/ntuple -> tag: MELA
        /common_path/SVFit/ntuple -> tag: SVFit
        Since when we compare two ntuples (with full path) only one
        directory changes in this scheme (see MELA vs SVFit), we
        create a list called 'tags' with these two strings; then we
        assign this string to friend.tag, if it's None
        """
        for f1, f2 in itertools.combinations(friends, 2):
            l1 = f1.path.split('/')
            l2 = f2.path.split('/')
            tags = list(set(l1).symmetric_difference(set(l2)))
            if tags:
                for t in tags:
                    if t in l1 and f1.tag is None:
                        f1.tag = t
                    elif t in l2 and f2.tag is None:
                        f2.tag = t
        return friends

    # E.g.: file_base_dir/file_name/file_name.root
    root_files = [os.path.join(files_base_directory, f, "{}.root".format(f)) for f in file_names]

    # E.g.: file_base_dir/file_name1/file_name1.root/folder/ntuple
    #       file_base_dir/file_name1/file_name2.root/folder/ntuple
    ntuples = []
    for root_file, file_name in zip(root_files, file_names):
        tdf_tree = get_full_tree_name(folder, root_file, 'ntuple')
        friends = []
        for friends_base_directory in friends_base_directories:
            friend_path = os.path.join(friends_base_directory, file_name, "{}.root".format(file_name))
            tdf_tree_friend = get_full_tree_name(folder, friend_path, 'ntuple')
            if tdf_tree != tdf_tree_friend:
                raise Exception("Extracted wrong TDirectoryFile from friend which is not the same than the base file.")
            friends.append(Ntuple(friend_path, tdf_tree_friend))
        ntuples.append(Ntuple(root_file, tdf_tree, add_tagged_friends(friends)))

    return Dataset(dataset_name, ntuples)


def dataset_from_files(dataset_name, tree_name, file_names):
    """Create a Dataset object from a list containing the names
    of the ROOT files (e.g. [root_file1, root_file2, (...)]):
    E.g.:
        my_dataset = dataset_from_files('my_dataset', 'tree_name', ['f1.root', 'f2.root'])

    Args:
        dataset_name (str): Name of the dataset
        tree_name (str): Name of the tree spanned through multiple files
        file_names (list): List containing the names of the .root
            files

    Returns:
        dataset (Dataset): Dataset object containing TTrees
    """
    def return_existent_tuple(file_name, tree_name):
        # Use TFile.Open() instead of TFile() in order to deal with
        # files accessed from remote
        root_file = TFile.Open(file_name)
        if root_file.IsZombie():
            raise FileNotFoundError('File {} does not exist, abort'.format(file_name))
        if tree_name not in root_file.GetListOfKeys():
            raise NameError('Tree {} does not exist in {}\n'.format(tree_name, file_name))
        root_file.Close()
        return Ntuple(file_name, tree_name)

    if not isinstance(file_names, list):
        raise TypeError('A list containing file names is required')

    ntuples = [return_existent_tuple(file_name, tree_name) for file_name in file_names]

    return Dataset(dataset_name, ntuples)


class Unit:
    """
    Building element of a minimal analysis flow, consisting
    of a dataset, a set of selections to apply on the data
    and a set of actions.

    Args:
        dataset (Dataset): Set of TTree objects to run the
            analysis on
        selections (list): List of Selection-type objects
        actions (Action): Actions to perform on the processed
            dataset, can be 'Histogram' or 'Count'
        variation (Variation): Variations applied, meaning
            that this selection is the result of a variation
            applied on other selections

    Attributes:
        dataset (Dataset): Set of TTree objects to run the
            analysis on
        selections (list): List of Selection-type objects
        actions (Action): Actions to perform on the processed
            dataset, can be 'Histogram' or 'Count'
        variation (Variation): Variations applied, meaning
            that this selection is the result of a variation
            applied on other selections
    """
    def __init__(
            self,
            dataset, selections, actions,
            variation = None):
        self.__set_dataset(dataset)
        self.__set_selections(selections)
        self.__set_actions(actions, variation)

    def __str__(self):
        layout = '\n'.join([
            'Dataset: {}'.format(self.dataset.name),
            'Selections: {}'.format(self.selections.__repr__()),
            'Actions: {}'.format(self.actions.__repr__())])
        return layout

    def __set_dataset(self, dataset):
        if not isinstance(dataset, Dataset):
            raise TypeError('not a Dataset object.')
        self.dataset = dataset

    def __set_selections(self, selections):
        if not isinstance(selections, list):
            raise TypeError('not a list object.')
        for selection in selections:
            if not isinstance(selection, Selection):
                raise TypeError('not a Selection object.')
        self.selections = selections

    def __set_actions(self, actions, variation):
        if not isinstance(actions, list):
            raise TypeError('not a list object.')
        for action in actions:
            if not isinstance(action, Action):
                raise TypeError('not an Action object.')
        self.actions = [self.__set_new_action(action, variation) \
                for action in actions]

    def __set_new_action(self, action, variation):
        if variation is None:
            name = '#'.join([self.dataset.name,
                '-'.join([selection.name for selection in self.selections]),
                action.name, 'Nominal'])
        else:
            if not isinstance(variation, Variation):
                raise TypeError('not a Variation object.')
            self.variation = variation
            name = action.name.replace('Nominal', self.variation.name)
        if isinstance(action, Histogram):
            if action.edges:
                return Histogram(name, action.variable, action.edges, action.expression)
            else:
                return Histogram(name, action.variable,
                        (action.nbins, action.low, action.up), action.expression)
        elif isinstance(action, Count):
            return Count(name, action.variable)

    def __eq__(self, other):
        return self.dataset == other.dataset and \
            self.selections == other.selections and \
            self.actions == other.actions

    def __hash__(self):
        return hash((
            self.dataset, tuple(self.selections),
            tuple(self.actions)))


class UnitManager:
    """
    Manager of all the Unit objects that are created.
    It can both be initialized with a variable amount of Unit
    objects as arguments or with no arguments, with the above mentioned
    objects added in a second time with the function 'book'.

    Attributes:
        booked_units (list): List of the booked units, updated during
            initialization or with the function 'book'
    """

    booked_units = []

    def book(self, units, variations = None):
        for unit in units:
            if unit not in self.booked_units:
                self.booked_units.append(unit)
        if variations:
            for variation in variations:
                logger.debug('Applying variation {}'.format(variation))
                for unit in units:
                    self.apply_variation(unit, variation)
        for action1, action2 in itertools.combinations([j for i in [
            unit.actions for unit in self.booked_units] for j in i], 2):
            if action1.name == action2.name:
                raise NameError('Caught two actions with same name ({}, {})'.format(
                    action1.name, action2.name))

    def apply_variation(self, unit, variation):
        new_unit = variation.create(unit)
        self.booked_units.append(new_unit)



from .booking import Ntuple
from .booking import Dataset
from .booking import Selection
from .booking import Histogram
from .booking import dataset_from_artusoutput
from .booking import dataset_from_files
from .booking import Unit
from .booking import UnitManager
from .optimization import GraphManager
from .run import RunManager
from .variations import ReplaceCut
from .inspect import get_dataframe



from copy import deepcopy
from collections import Counter

from .booking import Unit
from .utils import Node
from .utils import PrintedNode
from .utils import drawTree2

import logging
logger = logging.getLogger(__name__)



class Graph(Node):
    """
    A Graph is itself a node and every node has other nodes as children.
    Due to the way it is constructed, the root node will always be of
    kind 'dataset'.

    Args:
        unit (Unit): unit object from which a graph
            in its basic form is generated
        split_selections (Bool): boolean value
            indicating if we want to split the selections
            into minimal units

    Attributes:
        unit (Unit): unit object from which a graph
            in its basic form is generated
        split_selections (Bool): boolean value
            indicating if we want to split the selections
            into minimal units
    """
    def __init__(self, unit, split_selections = False):
        logger.debug('%%%%%%%%%% Constructing graph from Unit')
        self.split_selections = split_selections
        Node.__init__(self,
            unit.dataset.name,
            'dataset',
            unit.dataset)
        nodes = self.__nodes_from_unit(unit)
        if len(nodes) > 1:
            for no_last, no_first in zip(
                    nodes[:-1], nodes[1:]):
                if isinstance(no_first, Node):
                    no_last.children.append(no_first)
                elif isinstance(no_first, list):
                    for action in no_first:
                        no_last.children.append(action)
            self.children.append(nodes[0])
        else:
            # no-selections case: the only element of nodes is the
            # list of actions
            self.children.extend(nodes[0])

    def __nodes_from_unit(self, unit):
        nodes = list()
        last_node = list()
        for selection in unit.selections:
            if self.split_selections:
                sub_selections = selection.split()
                for sub_selection in sub_selections:
                    nodes.append(
                        Node(
                            sub_selection.name,
                            'selection',
                            sub_selection))
            else:
                nodes.append(
                    Node(
                        selection.name,
                        'selection',
                        selection))
        for action in unit.actions:
            last_node.append(
                Node(
                    action.name,
                    'action',
                    action))
        nodes.append(last_node)
        return nodes


class GraphManager:
    """
    Manager for Graph-type objects, with the main function of
    optimize/merge them with the 'optimize' function.

    Args:
        units (list): List of Unit objects used to
            fill the 'graphs' attribute

    Attributes:
        graphs (list): List of Graph objects that at some point
            will be merged and optimized
    """
    def __init__(self, units, split_selections = False):
        self.graphs = [
            Graph(unit, split_selections) for unit in units]

    def add_graph(self, graph):
        self.graphs.append(graph)

    def add_graph_from_unit(self, unit):
        self.graphs.append(Graph(unit))

    def optimize(self, level = 2):
        if int(level) == 0:
            logger.debug('No optimization selected.')
        elif int(level) == 1:
            logger.debug('Level 1 optimization selected: merge datasets.')
            self.merge_datasets()
        elif int(level) >= 2:
            logger.debug('Level 2 optimization selected: merge datasets and selections.')
            self.merge_datasets()
            self.optimize_selections()
        else:
            logger.debug('Invalid level of optimization, default to FULL OPTIMIZED.')
            self.merge_datasets()
            self.optimize_selections()
        logger.debug('Merged graphs:\n{}'.format(self.get_pretty_printed_merged_graphs()))

    def merge_datasets(self):
        logger.debug('%%%%%%%%%% Merging datasets:')
        merged_graphs = list()
        for graph in self.graphs:
            if graph not in merged_graphs:
                merged_graphs.append(graph)
            else:
                for merged_graph in merged_graphs:
                    if merged_graph == graph:
                        for child in graph.children:
                            merged_graph.children.append(child)
        self.graphs = merged_graphs
        logger.debug('%%%%%%%%%% Merging datasets: DONE')

    def optimize_selections(self):
        logger.debug('%%%%%%%%%% Optimizing selections:')
        for merged_graph in self.graphs:
            self._merge_children(merged_graph)
        logger.debug('%%%%%%%%%% Optimizing selections: DONE')

    def get_pretty_printed_merged_graphs(self):
        def call_node_rec(nd):
            return PrintedNode(nd.__repr__())([call_node_rec(child) for child in nd.children])
        fancy_trees = [call_node_rec(graph) for graph in self.graphs]
        return '\n\n'.join([drawTree2(False)(False)(t) for t in fancy_trees])

    def _merge_children(self, node):
        '''For every node, loops through the children
        and merges the ones that are equal, by appending
        the children of the new spotted ones to the
        children of the first spotted.
        '''
        merged_children = list()
        for child in node.children:
            if child not in merged_children:
                merged_children.append(child)
            else:
                merged_children[merged_children.index(
                    child)].children.extend(
                        child.children)
        node.children = merged_children
        for child in node.children:
            self._merge_children(child)



import logging
logger = logging.getLogger(__name__)



class Ntuple:
    def __init__(self, path, directory,
            friends = None, tag = None):
        self.path = path
        self.directory = directory
        if friends:
            self.friends = friends
        else:
            self.friends = list()
        self.tag = tag

    def __str__(self):
        if self.tag is None:
            layout = '({}, {})'.format(self.path, self.directory)
        else:
            layout = '({}, {}, tag = {})'.format(
                    self.path, self.directory, self.tag)
        return layout

    def __eq__(self, other):
        return self.path == other.path and \
            self.directory == other.directory

    def __hash__(self):
        return hash((
            self.path, self.directory))


class Dataset:
    def __init__(self, name, ntuples):
        self.name = name
        self.ntuples = ntuples

    def __str__(self):
        return 'Dataset-{}'.format(self.name)

    def __repr__(self):
        return self.__str__()

    def add_to_ntuples(self, *new_ntuples):
        for new_ntuple in new_ntuples:
            self.ntuples.append(new_ntuple)

    def __eq__(self, other):
        return self.name == other.name and \
            self.ntuples == other.ntuples

    def __hash__(self):
        return hash((
            self.name, tuple(self.ntuples)))


class Operation:
    def __init__(
            self, expression, name):
        self.expression = expression
        self.name = name

    def __eq__(self, other):
        return self.expression == other.expression and \
            self.name == other.name

    def __hash__(self):
        return hash((
            self.expression, self.name))


class Cut(Operation):
    def __str__(self):
        return 'Cut(' + self.name \
                + ', ' + self.expression \
                + ')'

    def __repr__(self):
        return self.__str__()


class Weight(Operation):
    def __str__(self):
        return 'Weight(' + self.name \
                + ', ' + self.expression \
                + ')'

    def __repr__(self):
        return self.__str__()

    def square(self):
        self.name = self.name + '^2'
        self.expression = '({0:})*({0:})'.format(self.expression)


class Selection:
    def __init__(
            self, name = None,
            cuts = None, weights = None):
        self.name = name
        self.set_cuts(cuts)
        self.set_weights(weights)

    def __str__(self):
        return 'Selection-{}'.format(self.name)

    def __repr__(self):
        return 'Selection-{}'.format(self.name)

    def __eq__(self, other):
        return self.cuts == other.cuts and \
            self.weights == other.weights

    def __hash__(self):
        return hash((tuple(self.cuts), tuple(self.weights)))

    def split(self):
        minimal_selections = list()
        for cut in self.cuts:
            s = Selection(name = '-'.join([cut.name, self.name]),
                    cuts = [cut])
            minimal_selections.append(s)
        for weight in self.weights:
            s = Selection(name = '-'.join([weight.name, self.name]),
                    weights = [weight])
            minimal_selections.append(s)
        return minimal_selections

    def add_cut(self, cut_expression, cut_name):
        self.cuts.append(Cut(
            cut_expression, cut_name))

    def add_weight(self, weight_expression, weight_name):
        self.weights.append(Weight(
            weight_expression, weight_name))

    def remove_cut(self, cut_name):
        for cut in self.cuts:
            if cut.name is cut_name:
                self.cuts.remove(cut)

    def remove_weight(self, weight_name):
        for weight in self.weights:
            if weight.name is weight_name:
                self.weights.remove(weight)

    def set_cuts(self, cuts):
        self.cuts = list()
        if cuts is not None:
            if isinstance(cuts, list):
                for cut in cuts:
                    if isinstance(cut, Cut):
                        self.cuts.append(cut)
                    elif isinstance(cut, tuple):
                        self.cuts.append(Cut(*cut))
                    else:
                        raise TypeError('not a Cut object or tuple')
            else:
                raise TypeError('a list is needed')

    def set_weights(self, weights):
        self.weights = list()
        if weights is not None:
            if isinstance(weights, list):
                for weight in weights:
                    if isinstance(weight, Weight):
                        self.weights.append(weight)
                    elif isinstance(weight, tuple):
                        self.weights.append(Weight(*weight))
                    else:
                        raise TypeError('not a Weight object or tuple')
            else:
                raise TypeError('a list is needed')


class Action:
    def __init__(self,
            name, variable):
        self.name = name
        self.variable = variable

    def __str__(self):
        return  self.name

    def __repr__(self):
        return  self.name

class Count(Action):
    pass


class Histogram(Action):
    def __init__(self, name, variable, setting, expression = None):
        Action.__init__(self, name, variable)
        if isinstance(setting, list):
            self.edges = setting
        elif isinstance(setting, tuple):
            self.nbins, self.low, self.up = setting
            self.edges = None
        else:
            raise TypeError('Argument {} must be either list or tuple'.format(setting))
        self.expression = expression

    def __eq__(self, other):
        if self.edges:
            return self.name == other.name and \
                self.variable == other.variable and \
                self.edges == other.edges
        else:
            return self.name == other.name and \
                self.variable == other.variable and \
                self.nbins == other.nbins and \
                self.low == other.low and \
                self.up == other.up

    def __hash__(self):
        if self.edges:
            return hash((self.name, self.variable, tuple(self.edges)))
        else:
            return hash((self.name, self.variable, self.nbins, self.low, self.up))



import unittest

from ntupro.booking import Ntuple, Dataset, Cut, Weight


class TestBookingMethods(unittest.TestCase):
    """ Test main classes and methods inside
    the booking submodule of ntuple_processor
    """
    def setUp(self):
        self.nt = Ntuple('path', 'directory')
        self.nt_friend = Ntuple('friend_path', 'friend_directory')
        self.ds = Dataset('ds', [self.nt])
        self.ct = Cut('cut_exp', 'cut_name')
        self.wh = Weight('weight_exp', 'weight_name')

    def test_ntuples(self):
        """
        Ntuple objects equal if have same path and directory
        """
        same_nt = Ntuple(self.nt.path, self.nt.directory)
        other_nt = Ntuple('other_path', self.nt.directory)
        self.assertEqual(self.nt, same_nt)
        self.assertNotEqual(self.nt, other_nt)

    def test_ntuple_with_friends(self):
        """
        Friends don't influence the comparison
        """
        same_nt = Ntuple(self.nt.path, self.nt.directory, [self.nt_friend])
        other_nt = Ntuple('other_path', self.nt.directory, [self.nt_friend])
        self.assertEqual(self.nt, same_nt)
        self.assertNotEqual(same_nt, other_nt)

    def test_datasets(self):
        """
        Dataset objects equal if have same name and list of ntuples
        """
        same_ds = Dataset(self.ds.name, [self.nt])
        other_ds = Dataset(self.ds.name, [self.nt_friend])
        self.assertEqual(self.ds, same_ds)
        self.assertNotEqual(self.ds, other_ds)

    def test_cuts_weights(self):
        """
        Cuts and Weights equal if same name and expression
        """
        same_ct = Cut(self.ct.expression, self.ct.name)
        other_ct = Cut(self.ct.expression, 'other_cut_name')
        same_wh = Weight(self.wh.expression, self.wh.name)
        other_wh = Weight(self.wh.expression, 'other_weight_name')
        self.assertEqual(self.ct, same_ct)
        self.assertNotEqual(self.ct, other_ct)
        self.assertEqual(self.wh, same_wh)
        self.assertNotEqual(self.wh, other_wh)


if __name__ == '__main__':
    unittest.main()



from ._booking import Ntuple
from ._booking import Dataset
from ._booking import Cut
from ._booking import Weight
from ._booking import Selection
from ._booking import Action
from ._booking import Count
from ._booking import Histogram

from ._optimization import Node

from ._run import RDataFrameCutWeight
from ._run import rdf_from_dataset_helper

from ._printing import Node as PrintedNode
from ._printing import drawTree2

from ._variations import Variation



import logging
logger = logging.getLogger(__name__)

class Node:
    def __init__(self,
            name, kind, unit_block, *children):
        self.name = name
        self.kind = kind
        self.unit_block = unit_block
        self.children = [
            child for child in children]

    def __str__(self):
        return '|Name: {}, Type: {}, Children: {}|'.format(
                self.name, self.kind, str(len(self.children)))

    def __repr__(self):
        return self.name

    def __eq__(self, other):
        return self.name == other.name and \
            self.kind == other.kind and \
            self.unit_block == other.unit_block

    def __hash__(self):
        return hash((
            self.name, self.kind, self.unit_block))




from multiprocessing import Pool
from time import time

from .utils import Count
from .utils import Histogram
from .utils import RDataFrameCutWeight
from .utils import rdf_from_dataset_helper

from ROOT import gROOT
gROOT.SetBatch(True)
from ROOT import RDataFrame
from ROOT import TFile
from ROOT import TChain
from ROOT import EnableImplicitMT
from ROOT.std import vector

import logging
logger = logging.getLogger(__name__)



class RunManager:
    """Convert the Graph-style language into PyROOT/RDataFrame
    language and schedule RDataFrame operations, like the
    following:
        Dataset()     -->   RDataFrame()
        Selection()   -->   Filter()
        Count()       -->   Sum()
        Histogram()   -->   Histo1D()

    Args:
        graphs (list): List of Graph objects that are converted
            node by node to RDataFrame operations

    Attributes:
        graphs (list): List of graphs to be processed
        tchains (list): List of TChains created, saved as attribute
            for the class in order to not let them go out of scope
        friend_tchains (list): List of friend TChains created,
            saved as attribute for the class in otder to not let
            them out of scope
    """
    def __init__(self, graphs):
        self.graphs = graphs
        self.tchains = list()
        self.friend_tchains = list()
        self.rcws = list()

    def _run_multiprocess(self, graph):
        start = time()
        ptrs = self.node_to_root(graph)
        logger.debug('%%%%%%%%%% Ready to produce a subset of {} shapes'.format(
            len(ptrs)))
        results = list()
        for ptr in ptrs:
            th = ptr.GetValue()
            results.append(th)
        # Sanity check: event loop run only once for each RDataFrame
        for rcw in self.rcws:
            loops = rcw.frame.GetNRuns()
            if loops != 1:
                logger.warning('Event loop run {} times'.format(loops))
        end = time()
        logger.debug('Event loop for graph {:} run in {:.2f} seconds'.format(
            repr(graph), end - start))
        return results

    def run_locally(self, output, nworkers = 1, nthreads = 1):
        """Save to file the histograms booked.

        Args:
            output (str): Name of the output .root file
            nworkers (int): number of slaves passed to the
                multiprocessing.Pool() function
            nthreads (int): number of threads passed to the
                EnableImplicitMT function
        """
        if not isinstance(nthreads, int):
            raise TypeError('wrong type for nthreads')
        if nthreads < 1:
            raise ValueError('nthreads has to be larger zero')
        self.nthreads = nthreads
        if not isinstance(nworkers, int):
            raise TypeError('wrong type for nworkers')
        if nworkers < 1:
            raise ValueError('nworkers has to be larger zero')
        logger.info('Start computing locally results of {} graphs using {} workers with {} thread(s) each'.format(
            len(self.graphs), nworkers, nthreads))
        start = time()
        pool = Pool(nworkers)
        final_results = list(pool.map(self._run_multiprocess, self.graphs))
        final_results = [j for i in final_results for j in i]
        end = time()
        logger.info('Finished computations in {} seconds'.format(int(end - start)))
        logger.info('Write {} results from {} graphs to file {}'.format(
            len(final_results), len(self.graphs), output))
        root_file = TFile(output, 'RECREATE')
        for op in final_results:
            op.Write()
        root_file.Close()

    def node_to_root(self, node, final_results = None, rcw = None):
        if final_results is None:
            final_results = list()
        if node.kind == 'dataset':
            logger.debug('%%%%%%%%%% node_to_root, converting to ROOT language the following dataset node\n{}'.format(
                node))
            result = self.__rdf_from_dataset(
                node.unit_block)
            if result not in self.rcws:
                self.rcws.append(result)
        elif node.kind == 'selection':
            if len(node.children) > 1:
                logger.debug('%%%%%%%%%% node_to_root, converting to ROOT language the following crossroad node\n{}'.format(
                    node))
            result = self.__cuts_and_weights_from_selection(
                rcw, node.unit_block)
        elif node.kind == 'action':
            logger.debug('%%%%%%%%%% node_to_root, converting to ROOT language the following action node\n{}'.format(
                node))
            if isinstance(node.unit_block, Count):
                result = self.__sum_from_count(
                    rcw, node.unit_block)
            elif isinstance(node.unit_block, Histogram):
                result = self.__histo1d_from_histo(
                    rcw, node.unit_block)
        if node.children:
            for child in node.children:
                self.node_to_root(child, final_results, result)
        else:
            final_results.append(result)
        return final_results

    def __rdf_from_dataset(self, dataset):
        chain, self.friend_tchains = rdf_from_dataset_helper(dataset)
        if self.nthreads != 1:
            EnableImplicitMT(self.nthreads)
        # Keep main chain alive
        self.tchains.append(chain)
        rdf = RDataFrame(chain)
        rcw = RDataFrameCutWeight(rdf)
        return rcw

    def __cuts_and_weights_from_selection(self, rcw, selection):
        l_cuts = [cut for cut in rcw.cuts]
        l_weights = [weight for weight in rcw.weights]
        for cut in selection.cuts:
            l_cuts.append(cut)
        for weight in selection.weights:
            l_weights.append(weight)
        l_rcw = RDataFrameCutWeight(rcw.frame, l_cuts, l_weights)
        return l_rcw

    def __sum_from_count(self, rdf, count):
        return rdf.Sum(count.variable)

    def __histo1d_from_histo(self, rcw, histogram):
        name = histogram.name
        var = histogram.variable
        edges = histogram.edges
        if edges:
            nbins = len(edges) - 1
        else:
            nbins = histogram.nbins
            low = histogram.low
            up = histogram.up
        var_expression = histogram.expression

        # Create column from expression if present in Histogram object
        if var_expression:
            rcw.frame = rcw.frame.Define(var, var_expression)

        # Create macro weight string from sub-weights applied
        # (saved earlier as rdf columns)
        weight_expression = '*'.join(['(' + weight.expression + ')' for weight in rcw.weights])

        # Create macro cut string from sub-cuts applied
        # (saved earlier as rdf columns)
        cut_expression = ' && '.join(['(' + cut.expression + ')' for cut in rcw.cuts])
        if cut_expression:
            rcw.frame = rcw.frame.Filter(cut_expression)

        # Create std::vector with the histogram edges
        if edges:
            l_edges = vector['double']()
            for edge in edges:
                l_edges.push_back(edge)

        if not weight_expression:
            logger.debug('%%%%%%%%%% Attaching histogram called {}'.format(name))
            if edges:
                histo = rcw.frame.Histo1D((
                        name, name, nbins, l_edges.data()),
                        var)
            else:
                histo = rcw.frame.Histo1D((
                        name, name, nbins, low, up),
                        var)
        else:
            weight_name = name.replace('#', '_')
            weight_name = weight_name.replace('-', '_')
            rcw.frame = rcw.frame.Define(weight_name, weight_expression)
            logger.debug('%%%%%%%%%% Attaching histogram called {}'.format(name))
            if edges:
                histo = rcw.frame.Histo1D((
                    name, name, nbins, l_edges.data()),
                    var, weight_name)
            else:
                histo = rcw.frame.Histo1D((
                    name, name, nbins, low, up),
                    var, weight_name)

        return histo



'''Textually visualized tree, with vertically-centered parent nodes'''

from functools import reduce
from itertools import (chain, takewhile)

'''
               ┌ Epsilon
               ├─── Zeta
       ┌─ Beta ┼──── Eta
       │       │         ┌───── Mu
       │       └── Theta ┤
 Alpha ┤                 └───── Nu
       ├ Gamma ────── Xi ─ Omicron
       │       ┌─── Iota
       └ Delta ┼── Kappa
               └─ Lambda
'''
# Tree style and algorithm inspired by the Haskell snippet at:
# https://doisinkidney.com/snippets/drawing-trees.html


# drawTree2 :: Bool -> Bool -> Tree a -> String
def drawTree2(blnCompact):
    '''Monospaced UTF8 left-to-right text tree in a
       compact or expanded format, with any lines
       containing no nodes optionally pruned out.
    '''
    def go(blnPruned, tree):
        # measured :: a -> (Int, String)
        def measured(x):
            '''Value of a tree node
               tupled with string length.
            '''
            s = ' ' + str(x) + ' '
            return len(s), s

        # lmrFromStrings :: [String] -> ([String], String, [String])
        def lmrFromStrings(xs):
            '''Lefts, Mid, Rights.'''
            i = len(xs) // 2
            ls, rs = xs[0:i], xs[i:]
            return ls, rs[0], rs[1:]

        # stringsFromLMR :: ([String], String, [String]) -> [String]
        def stringsFromLMR(lmr):
            ls, m, rs = lmr
            return ls + [m] + rs

        # fghOverLMR
        # :: (String -> String)
        # -> (String -> String)
        # -> (String -> String)
        # -> ([String], String, [String])
        # -> ([String], String, [String])
        def fghOverLMR(f, g, h):
            def go(lmr):
                ls, m, rs = lmr
                return (
                    [f(x) for x in ls],
                    g(m),
                    [h(x) for x in rs]
                )
            return lambda lmr: go(lmr)

        # leftPad :: Int -> String -> String
        def leftPad(n):
            return lambda s: (' ' * n) + s

        # treeFix :: (Char, Char, Char) -> ([String], String, [String])
        #                               ->  [String]
        def treeFix(l, m, r):
            def cfix(x):
                return lambda xs: x + xs
            return compose(stringsFromLMR)(
                fghOverLMR(cfix(l), cfix(m), cfix(r))
            )

        def lmrBuild(w, f):
            def go(wsTree):
                nChars, x = wsTree['root']
                _x = ('─' * (w - nChars)) + x
                xs = wsTree['nest']
                lng = len(xs)

                # linked :: String -> String
                def linked(s):
                    c = s[0]
                    t = s[1:]
                    return _x + '┬' + t if '┌' == c else (
                        _x + '┤' + t if '│' == c else (
                            _x + '┼' + t if '├' == c else (
                                _x + '┴' + t
                            )
                        )
                    )

                # LEAF ------------------------------------
                if 0 == lng:
                    return ([], _x, [])

                # SINGLE CHILD ----------------------------
                elif 1 == lng:
                    def lineLinked(z):
                        return _x + '─' + z
                    rightAligned = leftPad(1 + w)
                    return fghOverLMR(
                        rightAligned,
                        lineLinked,
                        rightAligned
                    )(f(xs[0]))

                # CHILDREN --------------------------------
                else:
                    rightAligned = leftPad(w)
                    lmrs = [f(x) for x in xs]
                    return fghOverLMR(
                        rightAligned,
                        linked,
                        rightAligned
                    )(
                        lmrFromStrings(
                            intercalate([] if blnCompact else ['│'])(
                                [treeFix(' ', '┌', '│')(lmrs[0])] + [
                                    treeFix('│', '├', '│')(x) for x
                                    in lmrs[1:-1]
                                ] + [treeFix('│', '└', ' ')(lmrs[-1])]
                            )
                        )
                    )
            return lambda wsTree: go(wsTree)

        measuredTree = fmapTree(measured)(tree)
        levelWidths = reduce(
            lambda a, xs: a + [max(x[0] for x in xs)],
            levels(measuredTree),
            []
        )
        treeLines = stringsFromLMR(
            foldr(lmrBuild)(None)(levelWidths)(
                measuredTree
            )
        )
        return [
            s for s in treeLines
            if any(c not in '│ ' for c in s)
        ] if (not blnCompact and blnPruned) else treeLines

    return lambda blnPruned: (
        lambda tree: '\n'.join(go(blnPruned, tree))
    )


# GENERIC -------------------------------------------------

# Node :: a -> [Tree a] -> Tree a
def Node(v):
    '''Contructor for a Tree node which connects a
       value of some kind to a list of zero or
       more child trees.
    '''
    return lambda xs: {'type': 'Tree', 'root': v, 'nest': xs}


# compose (<<<) :: (b -> c) -> (a -> b) -> a -> c
def compose(g):
    '''Right to left function composition.'''
    return lambda f: lambda x: g(f(x))


# concatMap :: (a -> [b]) -> [a] -> [b]
def concatMap(f):
    '''A concatenated list over which a function has been mapped.
       The list monad can be derived by using a function f which
       wraps its output in a list,
       (using an empty list to represent computational failure).
    '''
    return lambda xs: list(
        chain.from_iterable(map(f, xs))
    )


# fmapTree :: (a -> b) -> Tree a -> Tree b
def fmapTree(f):
    '''A new tree holding the results of
       applying f to each root in
       the existing tree.
    '''
    def go(x):
        return Node(f(x['root']))(
            [go(v) for v in x['nest']]
        )
    return lambda tree: go(tree)


# foldr :: (a -> b -> b) -> b -> [a] -> b
def foldr(f):
    '''Right to left reduction of a list,
       using the binary operator f, and
       starting with an initial accumulator value.
    '''
    def g(x, a):
        return f(a, x)
    return lambda acc: lambda xs: reduce(
        g, xs[::-1], acc
    )


# intercalate :: [a] -> [[a]] -> [a]
# intercalate :: String -> [String] -> String
def intercalate(x):
    '''The concatenation of xs
       interspersed with copies of x.
    '''
    return lambda xs: x.join(xs) if isinstance(x, str) else list(
        chain.from_iterable(
            reduce(lambda a, v: a + [x, v], xs[1:], [xs[0]])
        )
    ) if xs else []


# iterate :: (a -> a) -> a -> Gen [a]
def iterate(f):
    '''An infinite list of repeated
       applications of f to x.
    '''
    def go(x):
        v = x
        while True:
            yield v
            v = f(v)
    return lambda x: go(x)


# levels :: Tree a -> [[a]]
def levels(tree):
    '''A list of the nodes at each level of the tree.'''
    return list(
        map_(map_(root))(
            takewhile(
                bool,
                iterate(concatMap(nest))(
                    [tree]
                )
            )
        )
    )


# map :: (a -> b) -> [a] -> [b]
def map_(f):
    '''The list obtained by applying f
       to each element of xs.
    '''
    return lambda xs: list(map(f, xs))


# nest :: Tree a -> [Tree a]
def nest(t):
    '''Accessor function for children of tree node.'''
    return t['nest'] if 'nest' in t else None


# root :: Tree a -> a
def root(t):
    '''Accessor function for data of tree node.'''
    return t['root'] if 'root' in t else None



from ROOT import TChain

import logging
logger = logging.getLogger(__name__)

class RDataFrameCutWeight:
    def __init__(self,
            frame, cuts = [], weights = []):
        self.frame = frame
        self.cuts = cuts
        self.weights = weights

    def __str__(self):
        return str((
            self.frame,
            self.cuts, self.weights))

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.frame == other.frame and \
            self.cuts == other.cuts and \
            self.weights == other.weights

    def __hash__(self):
        return hash((
            self.frame, self.cuts, self.weights))

def rdf_from_dataset_helper(dataset):
    t_names = [ntuple.directory for ntuple in \
        dataset.ntuples]
    if len(set(t_names)) == 1:
        tree_name = t_names.pop()
    else:
        raise NameError(
            'Impossible to create RDataFrame with different tree names')
    chain = TChain()
    ftag_fchain = {}
    friend_tchains = []
    for ntuple in dataset.ntuples:
        chain.Add('{}/{}'.format(
            ntuple.path, ntuple.directory))
        for friend in ntuple.friends:
            if friend.tag not in ftag_fchain.keys():
                ftag_fchain[friend.tag] = TChain()
            ftag_fchain[friend.tag].Add('{}/{}'.format(
                friend.path, friend.directory))
    for ch in ftag_fchain.values():
        chain.AddFriend(ch)
        # Keep friend chains alive
        friend_tchains.append(ch)
    return (chain, friend_tchains)



from ROOT import RDataFrame
from .utils import rdf_from_dataset_helper

def get_dataframe(dataset):
    tchain, friend_tchains = rdf_from_dataset_helper(dataset)
    rdf = RDataFrame(tchain)
    setattr(rdf, 'tchain', tchain)
    setattr(rdf, 'friend_tchains', friend_tchains)
    return rdf



from copy import deepcopy

from .booking import Unit
from .booking import dataset_from_artusoutput
from .utils import Selection
from .utils import Variation

import logging
logger = logging.getLogger(__name__)



class ChangeDataset(Variation):
    """
    Variation that with the method create makes a deepcopy of
    the dataset inside the unit passed as argument and substitutes
    the directory attribute with folder_name.

    Args:
        name (str): name used to identify the instance of
            this class
        folder_name (str): part of the name of the TDirectoryFile
            in the new dataset following the prefix 'channel_' and
            preceding the suffix '/tree_name', e.g. 'NewFolder' in
            'mt_NewFolder/ntuple'

    Attributes:
        name (str): name used to identify the instance of
            this class
        folder_name (str): part of the name of the TDirectoryFile
            in the new dataset following the prefix 'channel_' and
            preceding the suffix '/tree_name', e.g. 'NewFolder' in
            'mt_NewFolder/ntuple'
    """
    def __init__(self,
            name, folder_name):
        Variation.__init__(self, name)
        self.folder_name = folder_name

    def create(self, unit):
        # A perfect copy of the daatset is exactly what
        # we need, thus using deepcopy is fine
        def change_folder(ntuple):
            folder, tree = ntuple.directory.split('/')
            ntuple.directory = '{}_{}/{}'.format(
                    folder.split('_')[0], self.folder_name, tree)
        new_dataset = deepcopy(unit.dataset)
        for ntuple in new_dataset.ntuples:
            change_folder(ntuple)
            for friend in ntuple.friends:
                change_folder(friend)
        return Unit(new_dataset, unit.selections, unit.actions, self)


class ReplaceCut(Variation):
    def __init__(self,
            name, replaced_name, cut):
        Variation.__init__(self, name)
        self.replaced_name = replaced_name
        self.cut = cut

    def create(self, unit):
        # Check that the name is present in at least one of the selections and raise an
        # error if not
        if not set([cut.name for selection in unit.selections for cut in selection.cuts \
                if cut.name == self.replaced_name]):
            raise NameError('Cut {} not found in any selection of this Unit'.format(self.replaced_name))
        new_selections = list()
        for selection in unit.selections:
            copy_cuts = list()
            for cut in selection.cuts:
                if cut.name == self.replaced_name:
                    logger.debug('Substitute {} with {} in selection {}'.format(
                        cut, self.cut, selection))
                    copy_cuts.append(self.cut)
                else:
                    copy_cuts.append(cut)
            new_selections.append(Selection(
                selection.name,
                copy_cuts,
                selection.weights))
        return Unit(unit.dataset, new_selections, unit.actions, self)


class ReplaceWeight(Variation):
    def __init__(self,
            name, replaced_name, weight):
        Variation.__init__(self, name)
        self.replaced_name = replaced_name
        self.weight = weight

    def create(self, unit):
        # Check that the name is present in at least one of the selections and raise an
        # error if not
        if not set([weight.name for selection in unit.selections for weight in selection.weights \
                if weight.name == self.replaced_name]):
            raise NameError('Weight {} not found in any selection of this Unit'.format(self.replaced_name))
        new_selections = list()
        for selection in unit.selections:
            copy_weights = list()
            for weight in selection.weights:
                if weight.name == self.replaced_name:
                    logger.debug('Substitute {} with {} in selection {}'.format(
                        weight, self.weight, selection))
                    copy_weights.append(self.weight)
                else:
                    copy_weights.append(weight)
            new_selections.append(Selection(
                selection.name,
                selection.cuts,
                copy_weights))
        return Unit(unit.dataset, new_selections, unit.actions, self)


class RemoveCut(Variation):
    def __init__(self,
            name, removed_name):
        Variation.__init__(self, name)
        self.removed_name = removed_name

    def create(self, unit):
        # Check that the name is present in at least one of the selections and raise an
        # error if not
        if not set([cut.name for selection in unit.selections for cut in selection.cuts \
                if cut.name == self.removed_name]):
            raise NameError('Cut {} not found in any selection of this Unit'.format(self.removed_name))
        new_selections = [selection for selection in unit.selections]
        for new_selection in new_selections:
            new_selection.remove_cut(self.removed_name)
        return Unit(unit.dataset, new_selections, unit.actions, self)

class RemoveWeight(Variation):
    def __init__(self,
            name, removed_name):
        Variation.__init__(self, name)
        self.removed_name = removed_name

    def create(self, unit):
        # Check that the name is present in at least one of the selections and raise an
        # error if not
        if not set([weight.name for selection in unit.selections for weight in selection.weights \
                if weight.name == self.removed_name]):
            raise NameError('Weight {} not found in any selection of this Unit'.format(self.removed_name))
        new_selections = [selection for selection in unit.selections]
        for new_selection in new_selections:
            new_selection.remove_weight(self.removed_name)
        return Unit(unit.dataset, new_selections, unit.actions, self)


class AddCut(Variation):
    def __init__(self,
            name, cut):
        Variation.__init__(self, name)
        self.cut = cut

    def create(self, unit):
        new_selections = [selection for selection in unit.selections]
        new_selections.append(Selection(
            name = self.cut.name, cuts = [self.cut]))
        return Unit(unit.dataset, new_selections, unit.actions, self)


class AddWeight(Variation):
    def __init__(self,
            name, weight):
        Variation.__init__(self, name)
        self.weight = weight

    def create(self, unit):
        new_selections = [selection for selection in unit.selections]
        new_selections.append(Selection(
            name = self.weight.name, weights = [self.weight]))
        return Unit(unit.dataset, new_selections, unit.actions, self)


class SquareWeight(Variation):
    def __init__(self,
            name, weight_name):
        Variation.__init__(self, name)
        self.weight_name = weight_name

    def create(self, unit):
        # Check that the name is present in at least one of the selections and raise an
        # error if not
        if not set([weight.name for selection in unit.selections for weight in selection.weights \
                if weight.name == self.weight_name]):
            raise NameError('Weight {} not found in any selection of this Unit'.format(self.weight_name))
        new_selections = [selection for selection in unit.selections]
        for new_selection in new_selections:
            for weight in new_selection.weights:
                if weight.name == self.name:
                    weight.square()
        return Unit(unit.dataset, new_selections, unit.actions, self)


class ReplaceCutAndAddWeight(Variation):
    def __init__(self,
            name, replaced_name, cut, weight):
        Variation.__init__(self, name)
        self.replace_cut = ReplaceCut(name, replaced_name, cut)
        self.add_weight = AddWeight(name, weight)

    def create(self, unit):
        unit = self.replace_cut.create(unit)
        return self.add_weight.create(unit)



import setuptools

readme = open("README.md", "r")
long_description = ''
for line in readme:
    if not line.startswith('<img'):
        long_description += line

setuptools.setup(
    name="ntupro",
    version="0.0.0",
    author="Massimiliano Galli",
    author_email="massimiliano.galli.95@gmail.com",
    description="Innovative package to optimize HEP analyses based on ROOT RDataFrame",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/maxgalli/ntupro",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



class Variation:
    """Base class for variations. The main method
    of this kind of class is 'create', here only
    declared and reimplemented in every derived class.
    It applies the Variation object to a Unit object
    and returns a new Unit object.

    Attributes:
        name (str): name assigned to the variation
    """
    def __init__(self,
            name):
        self.name = name

    def create(self, unit):
        pass

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name
